@@ -35,6 +35,8 @@ import (
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/bootfromvolume"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/floatingips"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/keypairs"
+	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/schedulerhints"
+	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/servergroups"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/flavors"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/servers"
 	"github.com/gophercloud/gophercloud/openstack/identity/v3/tokens"
@@ -394,7 +396,9 @@ func getImageID(is *InstanceService, imageName string) (string, error) {
 	}
 }
 
-// InstanceCreate creates a compute instance
+// InstanceCreate creates a compute instance.
+// If ServerGroupName is nonempty and no server group exists with that name,
+// then InstanceCreate creates a server group with that name.
 func (is *InstanceService) InstanceCreate(clusterName string, name string, clusterSpec *openstackconfigv1.OpenstackClusterProviderSpec, config *openstackconfigv1.OpenstackProviderSpec, cmd string, keyName string, configClient configclient.ConfigV1Interface) (instance *Instance, err error) {
 	if config == nil {
 		return nil, fmt.Errorf("create Options need be specified to create instace")
@@ -430,6 +434,7 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 	}
 	// Get all network UUIDs
 	var nets []ServerNetwork
+	netsWithoutAllowedAddressPairs := map[string]struct{}{}
 	for _, net := range config.Networks {
 		opts := networks.ListOpts(net.Filter)
 		opts.ID = net.UUID
@@ -438,6 +443,9 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 			return nil, err
 		}
 		for _, netID := range ids {
+			if net.NoAllowedAddressPairs {
+				netsWithoutAllowedAddressPairs[netID] = struct{}{}
+			}
 			if net.Subnets == nil {
 				nets = append(nets, ServerNetwork{
 					networkID: netID,
@@ -505,7 +513,12 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 		var port ports.Port
 		if len(portList) == 0 {
 			// create server port
-			port, err = CreatePort(is, name, net, &securityGroups, &allowedAddressPairs)
+			if _, ok := netsWithoutAllowedAddressPairs[net.networkID]; ok {
+				// create ports without address pairs
+				port, err = CreatePort(is, name, net, &securityGroups, &[]ports.AddressPair{})
+			} else {
+				port, err = CreatePort(is, name, net, &securityGroups, &allowedAddressPairs)
+			}
 			if err != nil {
 				return nil, fmt.Errorf("Failed to create port err: %v", err)
 			}
@@ -661,6 +674,59 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 
 	}
 
+	// The Machine spec accepts both a server group ID and a server group
+	// name. If both are present, assert that they are consistent, else
+	// fail. If only the name is present, create the server group.
+	//
+	// This block validates or populates config.ServerGroupID.
+	if config.ServerGroupName != "" {
+		existingServerGroups, err := getServerGroupsByName(is.computeClient, config.ServerGroupName)
+		if err != nil {
+			return nil, fmt.Errorf("retrieving existing server groups: %v", err)
+		}
+
+		if config.ServerGroupID == "" {
+			switch len(existingServerGroups) {
+			case 0:
+				sg, err := createServerGroup(is.computeClient, config.ServerGroupName)
+				if err != nil {
+					return nil, fmt.Errorf("creating the server group: %v", err)
+				}
+				config.ServerGroupID = sg.ID
+			case 1:
+				config.ServerGroupID = existingServerGroups[0].ID
+			default:
+				return nil, fmt.Errorf("multiple server groups found with the same ServerGroupName")
+			}
+		} else {
+			switch len(existingServerGroups) {
+			case 0:
+				return nil, fmt.Errorf("incompatible ServerGroupID and ServerGroupName")
+			default:
+				var found bool
+				for _, existingServerGroup := range existingServerGroups {
+					if existingServerGroup.ID == config.ServerGroupID {
+						found = true
+						break
+					}
+				}
+				if !found {
+					return nil, fmt.Errorf("incompatible ServerGroupID and ServerGroupName")
+				}
+			}
+		}
+	}
+
+	// If the spec sets a server group, then add scheduler hint
+	if config.ServerGroupID != "" {
+		serverCreateOpts = schedulerhints.CreateOptsExt{
+			CreateOptsBuilder: serverCreateOpts,
+			SchedulerHints: schedulerhints.SchedulerHints{
+				Group: config.ServerGroupID,
+			},
+		}
+	}
+
 	server, err := servers.Create(is.computeClient, keypairs.CreateOptsExt{
 		CreateOptsBuilder: serverCreateOpts,
 		KeyName:           keyName,
@@ -673,7 +739,43 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 	return serverToInstance(server), nil
 }
 
-func (is *InstanceService) InstanceDelete(id string) error {
+func createServerGroup(computeClient *gophercloud.ServiceClient, name string) (*servergroups.ServerGroup, error) {
+	// Microversion "2.15" is the first that supports "soft"-anti-affinity.
+	// Microversions starting from "2.64" accept policies as a string
+	// instead of an array.
+	defer func(microversion string) {
+		computeClient.Microversion = microversion
+	}(computeClient.Microversion)
+	computeClient.Microversion = "2.15"
+
+	return servergroups.Create(computeClient, &servergroups.CreateOpts{
+		Name:     name,
+		Policies: []string{"soft-anti-affinity"},
+	}).Extract()
+}
+
+func getServerGroupsByName(computeClient *gophercloud.ServiceClient, name string) ([]servergroups.ServerGroup, error) {
+	pages, err := servergroups.List(computeClient).AllPages()
+	if err != nil {
+		return nil, err
+	}
+
+	allServerGroups, err := servergroups.ExtractServerGroups(pages)
+	if err != nil {
+		return nil, err
+	}
+
+	serverGroups := make([]servergroups.ServerGroup, 0, len(allServerGroups))
+	for _, serverGroup := range allServerGroups {
+		if serverGroup.Name == name {
+			serverGroups = append(serverGroups, serverGroup)
+		}
+	}
+
+	return serverGroups, nil
+}
+
+func (is *InstanceService) deleteInstancePorts(id string) error {
 	// get instance port id
 	allInterfaces, err := attachinterfaces.List(is.computeClient, id).AllPages()
 	if err != nil {
@@ -736,6 +838,15 @@ func (is *InstanceService) InstanceDelete(id string) error {
 		}
 	}
 
+	return nil
+}
+
+func (is *InstanceService) InstanceDelete(id string) error {
+	err := is.deleteInstancePorts(id)
+	if err != nil {
+		klog.Warningf("Couldn't delete all instance %v ports: %v", id, err)
+	}
+
 	// delete instance
 	return servers.Delete(is.computeClient, id).ExtractErr()
 }
